/* horde-mpi.c
 * some functionality used by cmdqbf when using the HordeQBF version of
 * DepQBF.  Basically for stop_flag
 */

#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include "depqbf/qdpll.h"
#include "cmdqbf.h"

extern cmdqbfv cmdq;
extern masterv master;

/* caught an alarm, let's check for an interrupt from master */
/* if using HordeQBF's DepQBF, the MPI send from the master sets the interrupt flag.
 * if using 6.03, we set a small budget so it ends soon
 * if using Glucose, we set a small budget so it ends soon
 */
void mpihorde_alarmhandler(int sig)
{
	int flag;
	mprintf4(("%d: caught a SIGALRM, checking for interrupt request\n",
		  cmdq.rank));

	if (cmdq.stopreq == MPI_REQUEST_NULL)
	{
		printf("?? not critical but should not happen\n");
		exit(1); /* TODO not this, just interrupt since that's safe */
		return;
	}

	MPI_Test(&cmdq.stopreq,&flag,MPI_STATUS_IGNORE);
	if (flag)
	{
		/* reset for next interrupt request */
                MPI_Irecv(&cmdq.stopflag, 1, MPI_INT, MASTER, STOPFLAG,
			  MPI_COMM_WORLD, &cmdq.stopreq);
		if (cmdq.stopflag == 0) /* why? */
		{
			printf("%d: ?? stopflag 0? interrupt\n",cmdq.rank);
			return;
		}
		mprintf(("%d: got interrupt from master (%d)\n",cmdq.rank,
			 cmdq.stopflag));
		cmdq.stopflag = 2;
#ifndef HORDEQBF_DEPQBF
		if (!cmdq.is_sat)
		{
			char *ret;
			ret = qdpll_configure(cmdq.solver, "--max-secs=1");
			/* TODO: remove this ret bit since never hit */
			if (ret != NULL)
			{
				printf("interrupt ret %s\n", ret);
				free(ret);
			}
		}
#endif
		if (cmdq.is_sat)
			minisat_set_prop_budget(cmdq.sat_solver, 1);
	}
	else
	{
		signal(sig, mpihorde_alarmhandler);
		alarm(cmdq.alarmsecs);
	}
}

/* initialize SIGALRM handler and stop_flag */
void mpi_hordeworker_init(void)
{
	if (cmdq.stopflag == 3) /* initial time only */
		MPI_Irecv(&cmdq.stopflag, 1, MPI_INT, MASTER, STOPFLAG,
			  MPI_COMM_WORLD, &cmdq.stopreq);
	cmdq.stopflag = 0;
#ifdef HORDEQBF_DEPQBF
	if (!cmdq.is_sat)
		qdpll_set_interrupt_flag(cmdq.solver, &cmdq.stopflag);
#endif
	signal(SIGALRM, mpihorde_alarmhandler);
	alarm(cmdq.alarmsecs);
}

/* master wants to interrupt target. if target == MASTER, interrupt all workers
 */
const int mstopflag = 1;
void master_interrupt(int target)
{
	int i;

	if (target != 0)
	{
		if (master.interrupted[target] > 0)
		{
			mprintf(("M: avoiding interrupting %d again\n",target));
			return;
		}
		mprintf(("M: interrupting %d\n", target));		
		MPI_Send(&mstopflag,1,MPI_INT, target, STOPFLAG,MPI_COMM_WORLD);
		master.interrupted[target]++;
		return;
	}
	for (i=1; i<cmdq.size; i++)
	{
		/* really want everyone to stop, so interrupt everyone again*/
		mprintf(("M: interrupting %d\n", i));
		MPI_Send(&mstopflag, 1, MPI_INT, i, STOPFLAG, MPI_COMM_WORLD);
		master.interrupted[i]++;
	}
}

/* choose a random working worker, interrupt them, and return their rank */
int master_rand_interrupt(void)
{
	int i, r = rand()%(cmdq.size-1), target;
	r++;

	for (i=0; i<cmdq.size; i++)
	{
		target = (i+r)%cmdq.size;
		if (master.curjobs[target] != NULL)
		{
			mprintf(("M: random interrupt of %d to get work to split (size_jlist:%lu rem_jobs:%lu)\n",target,master.size_jlist,master.rem_jobs));
			master_interrupt(target);
			return target;
		}
	}
	mprintf2(("M: tried but failed to find a worker to interrupt for work?\n"));
	return 0;
}
