/* parser.c
 * the parser
 * intended to be indepedent of MPI and DepQBF
 */

#include "cmdqbf.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

extern masterv master;
extern cmdqbfv cmdq;

static void bind_freevars(int *);

/* the parser */
/* parsing qdimacs -- based on DepQBF's parse() */
#define PARSER_SKIP_SPACE_WHILE(c)                   \
  while (isspace (c))                                \
    c = fgetc (master.input);
#define PARSER_SKIP_SPACE_DO_WHILE(c)                \
  do                                                 \
    {                                                \
      c = getc (master.input);                       \
    }                                                \
  while (isspace (c));

void cmdqbf_getinput(void)
{
	qpref *qblock;
	long *vars;
	long **clauses;
	int *boundvar = NULL; /* boundvar[i] 1 if i is bound */
	int numvars, numclauses;
	int i, curv, size, curc=-1;
	int started_clauses = 0;
	int broken = 0;
	int preamble_found = 0;
	int curp = -1;
	int v;
	char c;

	while ((c = fgetc(master.input)) != EOF)
	{
		PARSER_SKIP_SPACE_WHILE(c);
		if (c==EOF)
			break;
		while (c == 'c') /* eat comments */
		{
			while ((c = fgetc(master.input)) != '\n' && c != EOF)
				;
			c = fgetc(master.input);
		}
		PARSER_SKIP_SPACE_WHILE(c);
		if (c==EOF)
			break;

		if (c == 'p')
		{
			if (preamble_found == 1)
			{
				broken = 1;
				break;
			}
			i = fscanf(master.input, " cnf %d %d",
				   &numvars, &numclauses);
			if (i != 2)
			{
				broken = 1;
				break;
			}
			if (numvars<0)
			{
				broken = 1;
				break;
			}
			if (numclauses<0)
			{
				broken = 1;
				break;
			}
			preamble_found = 1;
			master.num_vars = numvars;
			master.num_clauses = numclauses;
			master.clauses = malloc(sizeof(long*)*
						master.num_clauses);
			master.varfreq = calloc(numvars+1, sizeof(long));
			master.posvarfreq = calloc(numvars+1, sizeof(long));
			boundvar = calloc(numvars+1, sizeof(int));
			continue; /* to next line, should require newline */
		}

		if (preamble_found != 1)
		{
			broken = 1;
			break;
		}

		/* read quantifiers */
		if (c == 'e' || c == 'a')
		{
			if (started_clauses == 1)
			{       /* quantifiers must be in front */
				broken = 1;
				break;
			}
			if (c == 'a') /* has a universal quantifier */
				cmdq.is_sat = 0;
			if (master.size_p>0 && c==master.prefix[curp].quant)
			{
				/* disallowed by standard, but be liberal */
				/* same block as previous, just add to it */
				qblock = &(master.prefix[curp]);
				curv = qblock->num_vars;
				size = qblock->num_vars;
			}
			else /* add new block */
			{
				master.size_p++;
				qblock = realloc(master.prefix,
						 sizeof(qpref)*(master.size_p));
				if (qblock==NULL)
				{
					broken=1;
					break;
				}
				master.prefix = qblock;
				curp++;
				qblock = &(master.prefix[curp]);
				curv = 0;
				size = 0;
				qblock->vars = NULL;
				qblock->num_vars = 0;
				qblock->quant = c;
			}
			/* fill out qblock */
			while (fscanf(master.input, " %d", &v)==1)
			{
				if (v==0) /* end of qblock */
					break;
				if (v<0 || v>numvars)
				{
					broken = 1;
					break;
				}
				size++;
				vars = realloc(qblock->vars,
					       sizeof(long)*(size));
				if (vars == NULL)
				{
					broken = 1;
					break;
				}
				qblock->vars = vars;
				qblock->vars[curv++] = v;
				boundvar[v] = 1;
#if 0
				if (v>0)
				{
					master.posvarfreq[v]++;
					master.varfreq[v]++;
				}
				else
					master.varfreq[-1*v]++;
#endif
			}
			qblock->num_vars = size;
			if (v!=0 || broken==1)
			{
				broken = 1;
				break;
			}
			else /* read a quantifier line */
				continue;
		} /* end reading quantifier block */

		if (isdigit(c) || c=='-') /* started clause */
		{
			started_clauses = 1;
			ungetc(c, master.input);
			clauses = master.clauses;
			size = 16;
			curc++;
			if (curc>=master.num_clauses)
			{
				printf("too many clauses\n");
				broken = 1;
				break;
			}
			vars = malloc(sizeof(long)*size);
			if (vars == NULL)
			{
				broken = 1;
				break;
			}
			clauses[curc] = vars;
			curv = 0;

			/* read clause */
			while (fscanf(master.input, "%d", &v)==1)
			{
				PARSER_SKIP_SPACE_DO_WHILE(c);
				ungetc(c, master.input);
				if (v>numvars || (-1*v)>numvars)
				{
					broken = 1;
					break;
				}
				/* increase clause size */
				if (curv > size-1)
				{
					vars = realloc(clauses[curc],
						       sizeof(long)*(size+16));
					if (vars == NULL)
					{
						broken = 1;
						break;
					}
					clauses[curc] = vars;
					size+=16;
				}
				vars[curv++] = v;
				if (v == 0) /* finished clause */
					break;
				if (v>0)
				{
					master.posvarfreq[v]++;
					master.varfreq[v]++;
				}
				else
					master.varfreq[-1*v]++;
			}
			if (broken == 1 || v!=0)
			{
				printf("broken last var %d\n", v);
				broken = 1;
				break;
			}
			continue;
		}
	} /* getc */

	fclose(master.input);
	if (broken==1)
	{
		cmdq.cleanstart = 0;
		free(boundvar);
		return;
	}
	if (curc != master.num_clauses-1)
	{
		fprintf(stderr, "Expected %d clauses but only got %d\n",
		master.num_clauses, curc+1);
		master.num_clauses = curc+1;
	}
	bind_freevars(boundvar);
	free(boundvar);
	return;
}

/* boundvar[i] is 1 if i is bound.  If free variables occur,
 * add a new leading existential block containing them
 */
static void bind_freevars(int *boundvar)
{
	unsigned int i, count=0;
	qpref *qblock;

	for (i=1; i<=master.num_vars; i++)
		if (boundvar[i]==0 && master.varfreq[i]!=0)
			count++;

	if (count == 0) /* no such variables */
		return;

	mprintf(("M: %d free variables occur, adding leading existential block\n", count));

	qblock = realloc(master.prefix, sizeof(qpref)*(master.size_p + 1));
	if (qblock == NULL)
	{
		fprintf(stderr, "Error allocating memory for free variables : attempting to proceed anyways\n");
		return; /* this isn't needed for correct output */
	}

	master.prefix = qblock;
	master.size_p++;

	/* shift blocks to free leading one */
	for (i=master.size_p-1 ; i>0; i--)
		master.prefix[i] = master.prefix[i-1];

	/* create new leading block */
	qblock = &(master.prefix[0]);
	qblock->vars = malloc(sizeof(long) * count);
	qblock->num_vars = count;
	qblock->quant = 'e';
	count=0;
	for (i=1; i<=master.num_vars; i++)
		if (boundvar[i]==0 && master.varfreq[i]!=0)
			qblock->vars[count++] = i;
}

/* for debugging the parser */
void print_formula(void)
{
	unsigned int i, j;
	long *vars;

	printf("p cnf %d %d\n", master.num_vars, master.num_clauses);
	for (i=0; i<master.size_p; i++)
	{
		putchar(master.prefix[i].quant);
		vars = master.prefix[i].vars;
		for (j=0; j<master.prefix[i].num_vars; j++)
			printf(" %ld", vars[j]);
		printf(" 0\n");
	}
	for (i=0; i<master.num_clauses; i++)
	{
		vars = master.clauses[i];
		for (j=0; ; j++)
		{
			printf("%ld ", vars[j]);
			if (vars[j] == 0)
				break;
		}
		putchar('\n');
	}
	return;
}
